# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Import data from a standard .stat file.
-

    Args:
        _stat_file: A .stat file path on your system as a string.

    Returns:
        location: A Ladybug Location object describing the location data in the STAT file.
        ashrae_zone: The ASHRAE climate zone of the STAT file. Numbers in the zone
            denote average temperature (0 = Hottest; 8 = Coldest). Letters in
            the zone denote wetness (A = Humid; B = Dry; C = Marine)
        koppen_zone: The Koppen climate zone of the STAT file. The Koppen climate
            system uses vegetation as in indicator fo climate and combines
            average monthly temperatures, precipitation, and the seasonality
            of precipitation.
        clear_dir_norm_rad: The hourly "Clear Sky" Direct Normal Radiation in Wh/m2.
            Such clear sky radiation is typically used for sizing cooling
            systems. If monthly optical depths are found within the STAT
            file, these values will come from the Revised ASHARAE Clear Sky
            (Tau model). If no optical depths are found, they will come from
            the original ASHARE lear sky model.
        clear_diff_horiz_rad: The hourly "Clear Sky" Diffuse Horizontal Radiation
            in Wh/m2. Such clear sky radiation is typically used for sizing
            cooling systems. If monthly optical depths are found within the
            STAT file, these values will come from the Revised ASHARAE Clear Sky
            (Tau model). If no optical depths are found, they will come from
            the original ASHARE lear sky model.
        ann_heat_dday_996: A DesignDay object representing the annual 99.6%
            heating design day.
        ann_heat_dday_990: A DesignDay object representing the annual 99.0%
            heating design day.
        ann_cool_dday_004: A DesignDay object representing the annual 0.4%
            cooling design day.
        ann_cool_dday_010: A DesignDay object representing the annual 1.0%
            cooling design day.
        monthly_ddays_050: A list of 12 DesignDay objects representing monthly
            5.0% cooling design days.
        monthly_ddays_100: A list of 12 DesignDay objects representing monthly
            10.0% cooling design days.
        extreme_cold_week: A Ladybug AnalysisPeriod object representing the
            coldest week within the corresponding EPW.
        extreme_hot_week: A Ladybug AnalysisPeriod object representing the
            hottest week within the corresponding EPW.
        typical_weeks: A list of Ladybug AnalysisPeriod objects representing
            typical weeks within the corresponding EPW.
            The type of week can vary depending on the climate.
            _
            For mid and high lattitude climates with 4 seasons (eg. New York),
            these weeks are for each of the 4 seasons ordered as follows:
            Winter, Spring, Summer, Autumn
            _
            For low lattitude climates with wet/dry seasons (eg. Mumbai),
            these weeks might also include:
            Wet Season, Dry Season
            _
            For equitorial climates with no seasons (eg. Singapore),
            This output is usually a single week representing typical
            conditions of the entire year.
"""

ghenv.Component.Name = "LB Import STAT"
ghenv.Component.NickName = 'ImportSTAT'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '4'


try:
    from ladybug.stat import STAT
    from ladybug.wea import Wea
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    stat_obj = STAT(_stat_file)

    # output location and climate zone
    location = stat_obj.location
    ashrae_zone = stat_obj.ashrae_climate_zone
    koppen_zone = stat_obj.koppen_climate_zone

    # output clear sky radiation
    try:  # first see if we can get the values from monthly optical depths
        wea = Wea.from_stat_file(_stat_file)
    except:  # no optical data was found; use the original clear sky
        wea = Wea.from_ashrae_clear_sky(location)
    clear_dir_norm_rad = wea.direct_normal_irradiance
    clear_diff_horiz_rad = wea.diffuse_horizontal_irradiance

    # output design day objects
    ann_heat_dday_996 = stat_obj.annual_heating_design_day_996
    ann_heat_dday_990 = stat_obj.annual_heating_design_day_990
    ann_cool_dday_004 = stat_obj.annual_cooling_design_day_004
    ann_cool_dday_010 = stat_obj.annual_cooling_design_day_010
    monthly_ddays_050 = stat_obj.monthly_cooling_design_days_050
    monthly_ddays_100 = stat_obj.monthly_cooling_design_days_100

    # output extreme and typical weeks
    extreme_cold_week = stat_obj.extreme_cold_week
    extreme_hot_week = stat_obj.extreme_hot_week
    typical_weeks = []
    seasonal_wks = [stat_obj.typical_winter_week, stat_obj.typical_spring_week,
                   stat_obj.typical_summer_week, stat_obj.typical_autumn_week]
    for wk in seasonal_wks:
        if wk is not None:
            typical_weeks.append(wk)
    typical_weeks.extend(stat_obj.other_typical_weeks)


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Calculate Predicted Mean Vote (PMV).
-
PMV is a thermal comfort model for use on the interior of buildings
where a heating or cooling system is operational.
Note that, for naturally ventilated buildings, the Adaptive thermal
comfort model should be used.
-

    Args:
        _air_temp: Data Collection or individual value for air temperature in C.
        _mrt_: Data Collection or individual value of mean radiant temperature
            (MRT) in C. Default is the same as the air_temp.
        _rel_humid: Data Collection or individual value for relative humidity in %.
            Note that percent values are between 0 and 100.
        _air_speed_: Data Collection or individual value for air speed in m/s.
            Default is a very low speed of 0.1 m/s, which is typical of the
            room air speeds induced by HVAC systems.
        _met_rate_: Data Collection or individual value of metabolic rate in met.
            Default is set to 1.1 met for seated, typing. Typical values include
            the following.
            * 1 met = Metabolic rate of a resting seated person
            * 1.2 met = Metabolic rate of a standing person
            * 2.4 met = Metabolic rate of a person walking at ~1 m/s (2 mph)
        _clothing_: Data Collection or individual value of clothing insulation in clo.
            Default is set to 0.7 clo for long sleeve shirt and pants. Typical values
            include the following.
            * 1 clo = Three-piece suit
            * 0.5 clo = Shorts + T-shirt
            * 0 clo = No clothing
        pmv_par_: Optional comfort parameters from the "LB PMV Comfort Parameters"
            component to specify the criteria under which conditions are
            considered acceptable/comfortable. The default will assume a
            PPD threshold of 10% and no absolute humidity constraints.
        _run: Set to True to run the component.

    Returns:
        report: Reports, errors, warnings, etc.
        pmv: Predicted Mean Vote (PMV).
            _
            PMV is a seven-point scale from cold (-3) to hot (+3) that was used
            in comfort surveys of P.O. Fanger.
            _
            Each interger value of the scale indicates the following:
                * -3 = Cold
                * -2 = Cool
                * -1 = Slightly Cool
                *  0 = Neutral
                * +1 = Slightly Warm
                * +2 = Warm
                * +3 = Hot
        ppd: Percentage of People Dissatisfied (PPD).
            _
            Specifically, this is defined by the percent of people who would have
            a PMV beyond acceptable thresholds (typically <-0.5 and >+0.5).
            Note that, with the PMV model, the best possible PPD achievable is 5%
            and most standards aim to have a PPD below 10%.
        set: Standard Effective Temperature (SET) in Celcius.
            _
            These temperatures describe what the given input conditions "feel like"
            in relation to a standard environment of 50% relative humidity,
            <0.1 m/s average air speed, and mean radiant temperature equal to
            average air temperature, in which the total heat loss from the skin
            of an imaginary occupant with an activity level of 1.0 met and a
            clothing level of 0.6 clo is the same as that from a person in the
            actual environment.
        comfort: Integers noting whether the input conditions are acceptable
            according to the assigned comfort_parameter.
            _
            Values are one of the following:
                * 0 = uncomfortable
                * 1 = comfortable
        condition: Integers noting the thermal status of a subject according to
            the assigned comfort_parameter.
            _
            Values are one of the following:
                * -2 = too dry (but thermally neutral)
                * -1 = cold
                *  0 = netural
                * +1 = hot
                * +2 = too humid (but thermally neutral)
        heat_loss: A list of 6 terms for heat loss from the human energy
            balance calculation that underlies PMV. Valeus are in W.
            _
            The terms are ordered as follows:
                - Conduction
                - Sweating
                - Latent Respiration
                - Dry Respiration
                - Radiation
                - Convection
        comf_obj: A Python object containing all inputs and results of the
            analysis.  This can be plugged into components like the "Comfort
            Statistics" component to get further information.
"""

ghenv.Component.Name = 'LB PMV Comfort'
ghenv.Component.NickName = 'PMV'
ghenv.Component.Message = '1.3.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

try:
    from ladybug.psychrometrics import humid_ratio_from_db_rh
    from ladybug.datatype.temperature import Temperature
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_comfort.collection.pmv import PMV
    from ladybug_comfort.parameter.pmv import PMVParameter
    from ladybug_comfort.pmv import predicted_mean_vote
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def extract_collections(input_list):
    """Process inputs into collections and floats."""
    defaults = [None, _air_temp, 0.1, None, 1.1, 0.7]
    data_colls = []
    for i, input in enumerate(input_list):
        if input is None:
            input_list[i] = defaults[i]
        elif isinstance(input, BaseCollection):
            data_colls.append(input)
        else:
            try:
                input_list[i] = float(input)
            except ValueError as e:
                raise TypeError('input {} is not valid. Expected float or '
                                'DataCollection. Got {}'.format(input, type(input)))
    return input_list, data_colls

if all_required_inputs(ghenv.Component) and _run is True:
    # Process inputs and assign defaults.
    input_list = [_air_temp, _mrt_, _air_speed_, _rel_humid, _met_rate_, _clothing_]
    input, data_colls = extract_collections(input_list)
    pmv_par = pmv_par_ or PMVParameter()
    
    if data_colls == []:
        # The inputs are all individual values.
        pmv_result = predicted_mean_vote(input[0], float(input[1]), input[2],
                                         input[3], input[4], input[5], 0,
                                         pmv_par.still_air_threshold)
        pmv = pmv_result['pmv']
        ppd = pmv_result['ppd']
        set = pmv_result['set']
        hr = humid_ratio_from_db_rh(input[0], input[3])
        comfort = pmv_par.is_comfortable(pmv_result['ppd'], hr)
        condition = pmv_par.discomfort_reason(pmv_result['pmv'], pmv_result['ppd'])
        heat_loss = [pmv_result['heat_loss']['cond'],
                    pmv_result['heat_loss']['sweat'],
                    pmv_result['heat_loss']['res_l'],
                    pmv_result['heat_loss']['res_s'],
                    pmv_result['heat_loss']['rad'],
                    pmv_result['heat_loss']['conv']]
    else:
        # The inputs include Data Collections.
        if not isinstance(_air_temp, BaseCollection):
            _air_temp = data_colls[0].get_aligned_collection(
                float(_air_temp), Temperature(), 'C')
        
        comf_obj = PMV(_air_temp, _rel_humid, _mrt_, _air_speed_, _met_rate_,
            _clothing_, 0, pmv_par)
        pmv = comf_obj.predicted_mean_vote
        ppd = comf_obj.percentage_people_dissatisfied
        set = comf_obj.standard_effective_temperature
        comfort = comf_obj.is_comfortable
        condition = comf_obj.discomfort_reason
        heat_loss = [comf_obj.heat_loss_conduction,
                    comf_obj.heat_loss_sweating,
                    comf_obj.heat_loss_latent_respiration,
                    comf_obj.heat_loss_dry_respiration,
                    comf_obj.heat_loss_radiation,
                    comf_obj.heat_loss_convection]


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Calculate date information from an hour of the year.
_
Date information includes the month of the year, day of the month and the
hour + minute of the day.
-

    Args:
        _hoy: A number between 0 and 8759 for an hour of the year.

    Returns:
        month: The month of the year on which the input hoy falls.
        day: The day of the month on which the input hoy falls.
        hour: The hour of the day on which the input hoy falls.
        minute: The minute of the hour on which the input hoy falls.
        date: The input information as a human-readable date time.
        
"""

ghenv.Component.Name = "LB HOY to DateTime"
ghenv.Component.NickName = 'DateTime'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.dt import DateTime
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    month = []
    day = []
    hour = []
    minute = []
    date = []

    for hoy in _hoy:
        datetime = DateTime.from_hoy(hoy)
        month.append(datetime.month)
        day.append(datetime.day)
        hour.append(datetime.hour)
        minute.append(datetime.minute)
        date.append(datetime)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Convert a DataCollection of point-in-time values to its time-aggregated equivalent.
_
For example, if the collection has a Power data type in W, this method will
return a collection with an Energy data type in kWh.
-

    Args:
        _data: A houry, sub-hourly or daily data collection that can be aggregated
            over time to yield data of a different metric. (eg. a data collection
            of Power values in W).

    Returns:
        data_aggr: The data collection aggregated over time. (eg. a data collection
            of Energy values in kWh).
"""

ghenv.Component.Name = 'LB Time Aggregate'
ghenv.Component.NickName = 'Aggr'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.datacollection import HourlyDiscontinuousCollection, DailyCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    accept = (HourlyDiscontinuousCollection, DailyCollection)
    assert isinstance(_data, accept), '_data must be a an hourly ot daily data ' \
        'collection. Got {}.'.format(type(_data))
    if _data.header.data_type.time_aggregated_type is not None:
        data_aggr = _data.to_time_aggregated()
    else:
        data_aggr = _data


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Open a file in whatever program is associated with the file extension.
This can be used to open simulation files in particular applications (eg. opening
an OSM file in the OpenStudio Application).
-

    Args:
        _file_path: Full path to a file to be opened.
    
    Returns:
        report: Reports, errors, warnings, etc.
"""

ghenv.Component.Name = 'LB Open File'
ghenv.Component.NickName = 'OpenFile'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

import os
import subprocess

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check that the file exists
    assert os.path.exists(_file_path), \
        'No file was found at: {}'.format(_file_path)
    _file_path = _file_path.replace('/', '\\') if os.name == 'nt' else _file_path
    os.startfile(_file_path)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Visualize a sky matrix from the "LB Cumulative Sky Matrix" component as a colored
dome, subdivided into patches with a radiation value for each patch.
-

    Args:
        _sky_mtx: A Sky Matrix from the "LB Cumulative Sky Matrix" component, which
            describes the radiation coming from the various patches of the sky.
        _center_pt_: A point for the center of the dome. (Default: (0, 0, 0))
        _scale_: A number to set the scale of the sky dome. The default is 1,
            which corresponds to a radius of 100 meters in the current Rhino
            model's unit system.
        projection_: Optional text for the name of a projection to use from the sky
            dome hemisphere to the 2D plane. If None, a 3D sky dome will be drawn
            instead of a 2D one. (Default: None) Choose from the following:
                * Orthographic
                * Stereographic
        show_comp_: Boolean to indicate whether only one dome with total radiation
            should be displayed (False) or three domes with the solar radiation
            components (total, direct, and diffuse) should be shown. (Default: False).
        legend_par_: An optional LegendParameter object to change the display of the
            sky dome (Default: None).

    Returns:
        report: ...
        mesh:  A colored mesh representing the intensity of radiation for each of
            the sky patches within the sky dome.
        compass: A set of circles, lines and text objects that mark the cardinal
            directions in relation to the sky dome.
        legend: A legend showing the kWh/m2 values that correspond to the colors
            of the mesh.
        title: A text object for the title of the sunpath.
        patch_vecs: A list of vectors for each of the patches of the sky dome.
            All vectors are unit vectors and point from the center towards each
            of the patches. They can be used to construct visualizations of the
            rays used to perform radiation analysis.
        patch_values: Radiation values for each of the sky patches in kWh/m2. This
            will be one list if show_comp_ is "False" and a list of 3 lists (aka.
            a Data Tree) for total, direct, diffuse if show_comp_ is "True".
        mesh_values: Radiation values for each face of the dome mesh in kWh/m2. This
            can be used to post-process the radiation data and then regenerate
            the dome visualization using the mesh output from this component
            and the "LB Spatial Heatmap" component. Examples of useful post-
            processing include converting the units to something other than
            kWh/m2, inverting the +/- sign of radiation values depending on
            whether radiation is helpful or harmful to building thermal loads,
            etc. This will be one list if show_comp_ is "False" and a list of
            3 lists (aka. a Data Tree) for total, direct, diffuse if show_comp_
            is "True".
"""

ghenv.Component.Name = 'LB Sky Dome'
ghenv.Component.NickName = 'SkyDome'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import math

try:
    from ladybug_geometry.geometry2d.pointvector import Point2D
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
    from ladybug_geometry.geometry3d.mesh import Mesh3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.viewsphere import view_sphere
    from ladybug.compass import Compass
    from ladybug.graphic import GraphicContainer
    from ladybug.legend import LegendParameters
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_point3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_vector3d
    from ladybug_rhino.fromobjects import legend_objects, compass_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.grasshopper import all_required_inputs, \
        de_objectify_output, list_to_data_tree
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def draw_dome(dome_data, center, dome_name, legend_par):
    """Draw the dome mesh, compass, legend, and title for a sky dome.

    Args:
        dome_data: List of radiation values for the dome data
        center: Point3D for the center of the sun path.
        dome_name: Text for the dome name, which will appear in the title.
        legend_par: Legend parameters to be used for the dome

    Returns:
        dome_mesh: A colored mesh for the dome based on dome_data.
        dome_compass: A compass for the dome.
        dome_legend: A leend for the colored dome mesh.
        dome_title: A title for the dome.
        values: A list of radiation values that align with the dome_mesh faces.
    """
    # create the dome mesh and ensure patch values align with mesh faces
    if len(dome_data) == 145:  # tregenza sky
        lb_mesh = view_sphere.tregenza_dome_mesh_high_res.scale(radius)
        values = []  # high res dome has 3 x 3 faces per patch; we must convert
        tot_i = 0  # track the total number of patches converted
        for patch_i in view_sphere.TREGENZA_PATCHES_PER_ROW:
            row_vals = []
            for val in dome_data[tot_i:tot_i + patch_i]:
                row_vals.extend([val] * 3)
            for i in range(3):
                values.extend(row_vals)
            tot_i += patch_i
        values = values + [dome_data[-1]] * 18  # last patch has triangular faces
    else:  #reinhart sky
        lb_mesh = view_sphere.reinhart_dome_mesh.scale(radius)
        values = dome_data + [dome_data[-1]] * 11  # last patch has triangular faces

    # move and/or rotate the mesh as needed
    if north != 0:
        lb_mesh = lb_mesh.rotate_xy(math.radians(north), Point3D())
    if center != Point3D():
        lb_mesh = lb_mesh.move(Vector3D(center.x, center.y, center.z))

    # project the mesh if requested
    if projection_ is not None:
        if projection_.title() == 'Orthographic':
            pts = (Compass.point3d_to_orthographic(pt) for pt in lb_mesh.vertices)
        elif projection_.title() == 'Stereographic':
            pts = (Compass.point3d_to_stereographic(pt, radius, center)
                   for pt in lb_mesh.vertices)
        else:
            raise ValueError(
                'Projection type "{}" is not recognized.'.format(projection_))
        pts3d = tuple(Point3D(pt.x, pt.y, z) for pt in pts)
        lb_mesh = Mesh3D(pts3d, lb_mesh.faces)

    # output the dome visualization, including legend and compass
    move_fac = radius * 0.15
    min_pt = lb_mesh.min.move(Vector3D(-move_fac, -move_fac, 0))
    max_pt = lb_mesh.max.move(Vector3D(move_fac, move_fac, 0))
    graphic = GraphicContainer(values, min_pt, max_pt, legend_par)
    graphic.legend_parameters.title = 'kWh/m2'
    lb_mesh.colors = graphic.value_colors
    dome_mesh = from_mesh3d(lb_mesh)
    dome_legend = legend_objects(graphic.legend)
    dome_compass = compass_objects(
        Compass(radius, Point2D(center.x, center.y), north), z, None, projection_,
        graphic.legend_parameters.font)

    # construct a title from the metadata
    st, end = metadata[2], metadata[3]
    time_str = '{} - {}'.format(st, end) if st != end else st
    title_txt = '{} Radiation\n{}\n{}'.format(
        dome_name, time_str, '\n'.join([dat for dat in metadata[4:]]))
    dome_title = text_objects(title_txt, graphic.lower_title_location,
                              graphic.legend_parameters.text_height,
                              graphic.legend_parameters.font)

    return dome_mesh, dome_compass, dome_legend, dome_title, values


if all_required_inputs(ghenv.Component):
    # set defaults for global variables
    _scale_ = 1 if _scale_ is None else _scale_
    radius = (100 * _scale_) / conversion_to_meters()
    if _center_pt_ is not None:  # process the center point into a Point2D
        center_pt3d = to_point3d(_center_pt_)
        z = center_pt3d.z
    else:
        center_pt3d, z = Point3D(), 0

    # deconstruct the sky matrix and derive key data from it
    metadata, direct, diffuse = de_objectify_output(_sky_mtx)
    north = metadata[0]  # first item is the north angle
    sky_type = 1 if len(direct) == 145 else 2  # i for tregenza; 2 for reinhart
    total = [dirr + difr for dirr, difr in zip(direct, diffuse)] # total radiation

    # override the legend default min and max to make sense for domes
    l_par = legend_par_.duplicate() if legend_par_ is not None else LegendParameters()
    if l_par.min is None:
        l_par.min = 0
    if l_par.max is None:
        l_par.max = max(total)

    # output patch patch vectors
    patch_vecs_lb = view_sphere.tregenza_dome_vectors if len(total) == 145 \
        else view_sphere.reinhart_dome_vectors
    patch_vecs = [from_vector3d(vec) for vec in patch_vecs_lb]

    # create the dome meshes
    if not show_comp_:  # only create the total dome mesh
        mesh, compass, legend, title, mesh_values = \
            draw_dome(total, center_pt3d, 'Total', l_par)
        patch_values = total
    else:  # create domes for total, direct and diffuse
        # loop through the 3 radiation types and produce a dome
        mesh, compass, legend, title, mesh_values = [], [], [], [], []
        rad_types = ('Total', 'Direct', 'Diffuse')
        rad_data = (total, direct, diffuse)
        for dome_i in range(3):
            cent_pt = Point3D(center_pt3d.x + radius * 3 * dome_i,
                              center_pt3d.y, center_pt3d.z)
            dome_mesh, dome_compass, dome_legend, dome_title, dome_values = \
                draw_dome(rad_data[dome_i], cent_pt, rad_types[dome_i], l_par)
            mesh.append(dome_mesh)
            compass.extend(dome_compass)
            legend.extend(dome_legend)
            title.append(dome_title)
            mesh_values.append(dome_values)
        patch_values = list_to_data_tree(rad_data)
        mesh_values = list_to_data_tree(mesh_values)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Adjust the assumptions of the passive strategies that can be overalid on the
Psychrometric Chart using the "LB PMV Polygon" component. The default assumptions
of each of the strategies are as follows: 
_
Thermal Mass + Night Vent - The polygon represents the conditions under which
shaded, night-flushed thermal mass can keep occupants cool. By default, this
polygon assumes that temperatures can get as high as 12 C above the max temperature
of the comfort polygon as long temperatures 8 hours before the hot hour are
3.0 C lower than the max temperture of the comfort polygon. This parameter
component can be used to adjust these two temperature values and the number of
hours that the building keeps its "coolth".
_
Occupant Use of Fans - This polygon is made by assuming that an air speed of 1.0 m/s
is the maximum speed tolerable before papers start blowing around and conditions
become annoying to occupants. The polygon is determined by running a PMV model
with this fan air speed and the PMV inputs of the warmest comfort conditions.
This parameter component can be used to adjust this maximum acceptable air speed.
_
Capture Internal Heat - The polygon is made by assuming a minimum building balance
point of 12.8 C and any conditions that are warmer than that will keep occupants
comfortable (up to the comfort polygon). It is assumed that, above this building
balance temperature, the building is free-running and occupants are able to open
windows as they wish to keep conditions from overshooting the comfort polygon.
Note that the default balance temperature of 12.8 C is fairly low and assumes a
significant amount of internal heat from people, equipment. etc. Or the building 
as a well-insulated envelope to ensure what internal heat there is can leave the
building slowly. This parameter component can be used to adjust the balance
temperature.
_
Passive Solar Heating - The polygon represents the conditions under which
sun-exposed thermal mass can keep occupants warm in winter. By default, this
polygon assumes that temperatures can get as high as 12 C above the max temperature
of the comfort polygon as long temperatures 8 hours before the hot hour are
3.0 C lower than the max temperture of the comfort polygon. This parameter
component can be used to adjust these two temperature values and the number of
hours that the building keeps its "coolth".
-

    Args:
        _day_above_comf_: A number in degrees Celsius representing the maximum daily
            temperature above the comfort range which can still be counted in
            the "Mass + Night Vent" polygon. (Default: 12 C).
        _night_below_comf_: A number in degrees Celsius representing the minimum temperature
            below the maximum comfort polygon temperature that the outdoor
            temperature must drop at night in order to count towards the "Mass
            + Night Vent" polygon. (Default: 3C).
        _fan_air_speed_: The air speed around the occupants that the fans create in m/s.
            This is used to create the "Occupant Use of Fans" polygon. Note that
            values above 1 m/s tend to blow papers off desks. (Default: 1.0 m/3)
        _balance_temp_: The balance temperature of the building in Celsius when accounting
            for all internal heat. This is used to create the "Capture Internal
            Heat" polygon. This value must be greater or equal to 5 C (balance
            temperatures below 10 C are exceedingly rare) and it should be less
            than the coldest temperature of the merged comfort polygon in order
            to be meaningful. (Default: 12.8 C)
        _solar_heat_cap_: A number representing the amount of outdoor solar flux (W/m2)
            that is needed to raise the temperature of the theoretical building by
            1 degree Celsius. The lower this number, the more efficiently the
            space is able to absorb passive solar heat. The default assumes a
            relatively small passively solar heated zone without much mass. A
            higher number will be required the larger the space is and the more
            mass that it has. (Default: 50 W/m2)
        _time_constant_: A number that represents the amount of time in hours that a
            therortical building can passively maintain its temperature. This
            is used to determine how many hours a space can maintain a cool
            temperature after night flushing for the "Mass + Night Vent" polygon.
            It is also used to determine how many hours a space can store solar
            radiation for the "Passive Solar Heating" polygon. The default
            assumes a relatively well-isulated building with a thermal mass
            typical of most contemporary buildings. Higher mass buildings will
            be able to support a longer time constant. (Default: 8 hours).

    Returns:
        strategy_par: Passive strategy parameters that can be plugged into the "LB
            PMV Polygon" to adjust the assumptions of the passive strategy
            polygons.
"""

ghenv.Component.Name = 'LB Passive Strategy Parameters'
ghenv.Component.NickName = 'StrategyPar'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_rhino.grasshopper import give_warning, objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def check_strategy(value, name, default, max, min):
    """Check a strategy parameter to ensure it is correct."""
    if value is None:
        strategy_par.append(default)
    elif value <= max and value >= min:
        strategy_par.append(value)
    else:
        strategy_par.append(default)
        msg = '"{}" must be between {} and {}. Got {}.\nReverting to default ' \
            'value of {}'.format(name, min, max, value, default)
        print(msg)
        give_warning(ghenv.Component, msg)


#check and add each of the strategies
strategy_par = []
check_strategy(_day_above_comf_, '_day_above_comf_', 12.0, 30.0, 0.0)
check_strategy(_night_below_comf_, '_night_below_comf_', 3.0, 15.0, 0.0)
check_strategy(_fan_air_speed_, '_fan_air_speed_', 1.0, 10.0, 0.1)
check_strategy(_balance_temp_, '_balance_temp_', 12.8, 20.0, 5.0)
check_strategy(_solar_heat_cap_, '_solar_heat_cap_', 50.0, 1000.0, 1.0)
check_strategy(_time_constant_, '_time_constant_', 8, 48, 1)
strategy_par = objectify_output('Passive Strategy Parameters', strategy_par)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Get a Data Collection that is normalized by an area value.
_
Note that this component will raise a ValueError if the data type in the header
of the data collection is not normalizable to yeild a useful type. Also note
that a ZeroDivisionError will be raised if the input area is equal to 0.
-

    Args:
        _data: A Data Collection to be normalized by the input _area.
        area: A number representing area by which all of the data is normalized.
        _unit_: Text for the units that the area value is in. Acceptable inputs include
            'm2', 'ft2' and any other unit that is supported. (Default: m2).
    
    Returns:
        data: A Ladybug data collection object derived from the operation between
            the two data inputs.
"""

ghenv.Component.Name = 'LB Area Normalize'
ghenv.Component.NickName = 'Normalize'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    unit = _unit_ if _unit_ is not None else 'm2'
    data = _data.normalize_by_area(_area, unit)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Convert a value or list of values from one unit to another.
-

    Args:
        _values: Values to be converted from one unit type to another.
        _from_u: Text indicating the units of the input _values (eg. 'C')
        _to_u: Text indicating the units of the output values (eg. 'K')

    Returns:
        all_u: A text string indicating all possible units
            that can be plugged into _from_u and _to_u.
        values: The converted numerical values.
"""

ghenv.Component.Name = 'LB Unit Converter'
ghenv.Component.NickName = 'Units'
ghenv.Component.Message = '1.3.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    import ladybug.datatype
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


all_u = [': '.join([key, ', '.join(val)]) for key, val in ladybug.datatype.UNITS.items()]

if all_required_inputs(ghenv.Component):
    base_type = None
    for key in ladybug.datatype.UNITS:
        if _from_u in ladybug.datatype.UNITS[key]:
            base_type = ladybug.datatype.TYPESDICT[key]()
            break
    else:
        msg = 'Input _from_u "{}" is not recgonized as a valid unit.\n Check all_u ' \
        '(with nothing connected to the component) to see the acceptable units.'.format(_from_u)
        raise ValueError(msg)

    values = base_type.to_unit(_values, _to_u, _from_u)


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Draw a psychrometric chart in the Rhino scene and plot a set of temperatures and
humidity values on it.
_
Connected data can be either outdoor temperature and humidty from imported EPW
weather data or indoor temperature and humidity ratios from an energy simulation.
-

    Args:
        _temperature: A hourly, daily, or sub-hourly data collection of temperature values
            or a single temperature value in Celcius to be used for the whole
            analysis. If this input data collection is in in Farenheit, the
            entire chart will be drawn using IP units. Operative temperature
            should be used if it is available. Otherwise, air temperature
            (aka. dry bulb temperature) is suitable.
        _rel_humidity: A hourly, daily, or sub-hourly data collection of relative humidity
            values in % or a single relative humidity value to be used for the
            whole analysis. Note that the input data collection here must align
            with the _temperature input.
        _pressure_: A data collection of atmospheric pressure in Pascals or a single
            number for the average air pressure across the data plotted on the
            chart. It is recommended that the barometric pressure from the
            "Import EPW" component be used here as the default is not sutiable
            for higher elevations. (Default: 101325 Pa; pressure at sea level).
        _base_pt_: A point to be used as the bottom-left-most point from which all
            geometry of the plot will be generated. (Default: (0, 0, 0)).
        _scale_: A number to set the dimensions of the chart. (Default: 1).
        _temp_range_: An optional domain (or number for the upper temperature), which
            will be used to set the lower and upper boundaries of temperature
            on the psychrometric chart. (Default: -20 to 55 when the chart is
            in SI; -5 to 115 when the chart is in IP).
        plot_wet_bulb_: Boolean to note whether the psychrometric chart should be ploted
            with lines of constant enthalpy (False) or lines of constant wet bulb
            temperature (True).  (Default: False).
        legend_par_: An optional LegendParameter object from the "LB Legend Parameters"
            component to change the display of the Pyschrometric Chart.
        data_: Optional data collections, which are aligned with the input _temperature
            and _rel_humidity, which will be output from the data of this
            component and can be used to color points with data. This data
            can also be used along with the statement_ below to select out
            data that meets certain conditions.
        statement_: A conditional statement as a string (e.g. a > 25).
            .   
            The variable of the first data collection should always be named 'a'
            (without quotations), the variable of the second list should be
            named 'b', and so on.
            .
            For example, if three data collections are connected to _data
            and the following statement is applied:
            '18 < a < 26 and b < 80 and c > 2'
            The resulting collections will only include values where the first
            data collection is between 18 and 26, the second collection is less
            than 80 and the third collection is greater than 2.
            .
            For this component, temperature will always be the second-to-last
            letter and relative humidity will be the last letter.
        period_: A Ladybug analysis period to be applied to the _temperature and
            _rel_humidity of the input data_.

    Returns:
        report: ...
        title: Text objects for the chart title and axes titles as well as a polyline for
            the outer border of the chart. Note that the polyline for the border
            excludes the saturation line.
        temp_lines: A list of line segments and text objects for the temperature labels
            on the chart.
        rh_lines: A list of curves and text objects for the relative humidity labels
            on the chart.
        hr_lines: A list of line segments and text objects for the humidty ratio labels
            on the chart.
        enth_wb_lines: A list of line segments and text objects for the enthalpy or
            wet bulb temperature labels on the chart.
        mesh: A colored mesh showing the number of input hours that happen in each part
            of the psychrometric chart.
        legend: A colored legend showing the number of hours that correspond to
            each color.
        points: Points representing each of the input temperature and humidity values.
            By default, this ouput is hidden and it should be connected it to a
            native Grasshopper preview component to view it.
        data: The input data_ with the input statements or the periods applied to it.
            These can be deconstructed with the "LB Deconstruct Data" component
            and the resulting values can be plugged into the "LB Create Legend"
            component to generate colors that can be used to color the points
            above using the native Grasshopper "Custom Preview" component.
        psych_chart: A Psychrometric Chart object, which can be connected to any
            of the "Comfort Polygon" components in order to plot polygons on
            the chart and perform thermal comfort analyses on the data.
"""

ghenv.Component.Name = 'LB Psychrometric Chart'
ghenv.Component.NickName = 'PsychrometricChart'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.datacollection import BaseCollection
    from ladybug.psychchart import PsychrometricChart
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
    from ladybug_geometry.geometry3d.plane import Plane
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_point2d, to_point3d
    from ladybug_rhino.fromgeometry import from_mesh2d, from_polyline2d, \
        from_linesegment2d, from_point2d
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.color import color_to_color
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree, \
        hide_output, show_output, longest_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def leg_par_by_index(i):
    """Get legend parameters associated with a given index of the chart."""
    try:
        return legend_par_[i]
    except IndexError:
        return None


def small_labels(psy_chart, labels, points, x_align, y_align, factor=1.0):
    """Translate a list of psych chart text labels into the Rhino scene."""
    return [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, z)),
                         psy_chart.legend_parameters.text_height * factor,
                         psy_chart.legend_parameters.font, x_align, y_align)
            for txt, pt in zip(labels, points)]


def plane_from_point(point_2d, align_vec=Vector3D(1, 0, 0)):
    """Get a Plane from a Point2D.

    Args:
        point_2d: A Point2D to serve as the origin of the plane.
        align_vec: A Vector3D to serve as the X-Axis of the plane.
    """
    return Plane(o=Point3D(point_2d.x, point_2d.y, z), x=align_vec)


def draw_psych_chart(psy_chart):
    """Draw a given psychrometric chart object into Rhino geometry.

    This will NOT translate any colored meshes or data points.
    """
    # output all of the lines/polylines for the various axes
    title_i = [from_polyline2d(psy_chart.chart_border, z)]
    temp_lines_i = [from_linesegment2d(tl, z) for tl in psy_chart.temperature_lines]
    rh_lines_i = [from_polyline2d(rhl, z) for rhl in psy_chart.rh_lines]
    hr_lines_i = [from_linesegment2d(hrl, z) for hrl in psy_chart.hr_lines]

    # add the text to the various lines
    title_i.append(text_objects(
        psy_chart.x_axis_text, plane_from_point(psy_chart.x_axis_location),
        psy_chart.legend_parameters.text_height * 1.5,
        psy_chart.legend_parameters.font, 0, 0))
    title_i.append(text_objects(
        psy_chart.y_axis_text, plane_from_point(psy_chart.y_axis_location, Vector3D(0, 1)),
        psy_chart.legend_parameters.text_height * 1.5,
        psy_chart.legend_parameters.font, 2, 0))
    temp_lines_i = temp_lines_i + small_labels(
        psy_chart, psy_chart.temperature_labels, psy_chart.temperature_label_points, 1, 0)
    rh_lines_i = rh_lines_i + small_labels(
        psy_chart, psy_chart.rh_labels[:-1], psy_chart.rh_label_points[:-1], 2, 3, 0.8)
    hr_lines_i = hr_lines_i + small_labels(
        psy_chart, psy_chart.hr_labels, psy_chart.hr_label_points, 0, 3)

    # add enthalpy or wet bulb lines
    if plot_wet_bulb_:
        enth_wb_lines_i = [from_linesegment2d(el, z) for el in psy_chart.wb_lines]
        enth_wb_lines_i = enth_wb_lines_i + small_labels(
            psy_chart, psy_chart.wb_labels, psy_chart.wb_label_points, 2, 3)
    else:
        enth_wb_lines_i = [from_linesegment2d(el, z) for el in psy_chart.enthalpy_lines]
        enth_wb_lines_i = enth_wb_lines_i + small_labels(
            psy_chart, psy_chart.enthalpy_labels, psy_chart.enthalpy_label_points, 2, 3)

    # add all of the objects to the bse list
    title.append(title_i)
    temp_lines.append(temp_lines_i)
    rh_lines.append(rh_lines_i)
    hr_lines.append(hr_lines_i)
    enth_wb_lines.append(enth_wb_lines_i)


if all_required_inputs(ghenv.Component):
    # process the base point
    bp = to_point2d(_base_pt_) if _base_pt_ is not None else Point2D()
    z = to_point3d(_base_pt_).z if _base_pt_ is not None else 0

    # create lists to be filled with objects
    title = []
    temp_lines = []
    rh_lines = []
    hr_lines = []
    enth_wb_lines = []
    mesh = []
    legend = []
    points = []
    data_colls = []
    psych_chart = []

    # loop through the input temperatures and humidity and plot psych charts
    for j, (temperature, rel_humid) in enumerate(zip(_temperature, _rel_humidity)):
        # process the pressure input
        pressure = 101325
        if len(_pressure_) != 0:
            pr = longest_list(_pressure_, j)
            try:
                pressure = float(pr)
            except Exception:  # assume that it's a data collection
                assert pr.header.unit == 'Pa', '_pressure_ input must be in Pa.'
                pressure = pr.average

        # sense if the input temperature is in Farenheit
        use_ip = False
        if isinstance(temperature, BaseCollection):
            if temperature.header.unit != 'C':  # convert to C and set chart to use_ip
                temperature = temperature.to_si()
                use_ip = True

        # set default values for the chart dimensions
        _scale_ = 1.0 if _scale_ is None else _scale_
        x_dim = _scale_ * 2 / conversion_to_meters()
        y_dim = (_scale_ * 2 * 1500) / conversion_to_meters()
        y_dim = y_dim * (9 / 5) if use_ip else y_dim
        t_min, t_max = (-5, 115) if use_ip else (-20, 50)
        if _temp_range_ is not None:
            t_min, t_max = _temp_range_
        y_move_dist = -y_dim * 0.04 * j
        base_pt = bp.move(Vector2D(0, y_move_dist))

        # apply any analysis periods and conditional statements to the input collections
        original_temperature = temperature
        all_data = data_ + [temperature, rel_humid]
        if period_ is not None:
            all_data = [coll.filter_by_analysis_period(period_) for coll in all_data]
        if statement_ is not None:
            all_data = BaseCollection.filter_collections_by_statement(all_data, statement_)

        # create the psychrometric chart object and draw it in the Rhino scene
        psy_chart = PsychrometricChart(
            all_data[-2], all_data[-1], pressure, leg_par_by_index(0), base_pt, x_dim, y_dim,
            t_min, t_max, use_ip=use_ip)
        psy_chart.z = z
        psy_chart.original_temperature = original_temperature
        draw_psych_chart(psy_chart)
        if isinstance(all_data[-2], BaseCollection):
            meta_i = all_data[-2].header.metadata.items()
            title_items = ['Time [hr]'] + ['{}: {}'.format(k, v) for k, v in meta_i]
        else:
            title_items = ['Psychrometric Chart']
        title[j + j * len(data_)].append(text_objects(
            '\n'.join(title_items), psy_chart.container.upper_title_location,
            psy_chart.legend_parameters.text_height * 1.5,
            psy_chart.legend_parameters.font, 0, 0))
        psych_chart.append(psy_chart)

        # plot the data on the chart
        lb_points = psy_chart.data_points
        points.append([from_point2d(pt) for pt in lb_points])
        if len(lb_points) != 1:  # hide the points and just display the mesh
            hide_output(ghenv.Component, 8)
            mesh.append(from_mesh2d(psy_chart.colored_mesh, z))
            legend.append(legend_objects(psy_chart.legend))
        else:  # show the single point on the chart
            show_output(ghenv.Component, 8)

        # process any of the connected data into a legend and colors
        if len(data_) != 0:
            data_colls.append(all_data[:-2])
            move_dist = x_dim * (psy_chart.max_temperature - psy_chart.min_temperature + 20)
            for i, d in enumerate(all_data[:-2]):
                # create a new psychrometric chart offset from the original
                new_pt = Point2D(base_pt.x + move_dist * (i + 1), base_pt.y)
                psy_chart = PsychrometricChart(
                    all_data[-2], all_data[-1], pressure, leg_par_by_index(0),
                    new_pt, x_dim, y_dim, t_min, t_max, use_ip=use_ip)
                psy_chart.z = z
                psy_chart.original_temperature = original_temperature
                draw_psych_chart(psy_chart)
                psych_chart.append(psy_chart)
                lb_mesh, container = psy_chart.data_mesh(d, leg_par_by_index(i + 1))
                mesh.append(from_mesh2d(lb_mesh, z))
                legend.append(legend_objects(container.legend))
                move_vec = Vector2D(base_pt.x + move_dist * (i + 1), 0)
                points.append([from_point2d(pt.move(move_vec)) for pt in lb_points])

                # add a title for the new chart
                title_items = ['{} [{}]'.format(d.header.data_type, d.header.unit)] + \
                    ['{}: {}'.format(key, val) for key, val in d.header.metadata.items()]
                title[j + j * len(data_) + (i + 1)].append(text_objects(
                    '\n'.join(title_items), psy_chart.container.upper_title_location,
                    psy_chart.legend_parameters.text_height * 1.5,
                    psy_chart.legend_parameters.font, 0, 0))

    # upack all of the python matrices into data trees
    title = list_to_data_tree(title)
    temp_lines = list_to_data_tree(temp_lines)
    rh_lines = list_to_data_tree(rh_lines)
    hr_lines = list_to_data_tree(hr_lines)
    enth_wb_lines = list_to_data_tree(enth_wb_lines)
    legend = list_to_data_tree(legend)
    points = list_to_data_tree(points)
    data = list_to_data_tree(data_colls)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>


"""
Dump any Ladybug data collections into a file. You can use "LB Load Data" component
to load the data collections from the file back into Grasshopper.
-

    Args:
        _data: A list of Ladybug data collections to be written to a file.
        _format_: Text or an integer to set the format of the output file.
            Choose from the options below. (Default: CSV).
                * 0 = CSV - Compact, human-readable, importable to spreadsheets
                * 1 = JSON - Cross-language and handles any types of collections
                * 2 = PKL - Compressed format only readable with Python
        _name_: A name for the file to which the data collections will be
            written. (Default: 'data').
        _folder_: An optional directory into which the data collections will be
            written.  The default is set to the simulation folder.
        _dump: Set to "True" to save the honeybee objects to file.

    Returns:
        report: Errors, warnings, etc.
        data_file: The path of the file where the data collections are saved.
"""

ghenv.Component.Name = 'LB Dump Data'
ghenv.Component.NickName = 'DumpData'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

import os

try:
    from ladybug.datautil import collections_to_csv, collections_to_json, \
        collections_to_pkl
    from ladybug.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import the core ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

FORMAT_MAP = {
    '0': 'csv',
    '1': 'json',
    '2': 'pkl',
    'csv': 'csv',
    'json': 'json',
    'pkl': 'pkl'
}


if all_required_inputs(ghenv.Component) and _dump:
    # set the component defaults
    name = _name_ if _name_ is not None else 'data'
    folder = _folder_ if _folder_ is not None else \
        os.path.join(os.path.dirname(folders.ladybug_tools_folder), 'simulation')
    format = 'csv' if _format_ is None else FORMAT_MAP[_format_.lower()]

    # write the data into the appropriate format
    if format == 'csv':
        try:
            data_file = collections_to_csv(_data, folder, name)
        except AssertionError as e:
            raise ValueError('{}\nTry using the JSON or PKL format.'.format(e))
    elif format == 'json':
        data_file = collections_to_json(_data, folder, name)
    elif format == 'pkl':
        data_file = collections_to_pkl(_data, folder, name)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Compute the hourly solar irradiance or illuminance falling on an unobstructed surface
that faces any direction.
_
The calculation method of this component is faster than running "LB Incident
Radiation" studies on an hour-by-hour basis and it is slighty more realistic since
it accounts for ground reflection. However, this comes at the cost of not being
able to account for any obstructions that block the sun.
-

    Args:
        _location: A Ladybug Location object, used to determine the altitude and
            azimuth of the sun at each hour.
        _direct_norm: Hourly Data Collection with the direct normal solar
            irradiance in W/m2 or Illuminance in lux.
        _diffuse_horiz: Hourly Data Collection with diffuse horizontal solar
            irradiance in W/m2 or Illuminance in lux.
        _srf_azimuth_: A number between 0 and 360 that represents the azimuth at which
            irradiance is being evaluated in degrees.  0 = North, 90 = East,
            180 = South, and 270 = West.  (Default: 180).
        _srf_altitude_: A number between -90 and 90 that represents the altitude at which
            irradiance is being evaluated in degrees. A value of 0 means the
            surface is facing the horizon and a value of 90 means a surface is
            facing straight up. (Default: 0).
        _ground_ref_: A number between 0 and 1 that represents the reflectance of the
            ground. (Default: 0.2). Some common ground reflectances are:
                *   urban: 0.18
                *   grass: 0.20
                *   fresh grass: 0.26
                *   soil: 0.17
                *   sand: 0.40
                *   snow: 0.65
                *   fresh_snow: 0.75
                *   asphalt: 0.12
                *   concrete: 0.30
                *   sea: 0.06
        anisotrophic_: A boolean value that sets whether an anisotropic sky is used
            (as opposed to an isotropic sky). An isotrophic sky assumes an
            even distribution of diffuse irradiance across the sky while an
            anisotropic sky places more diffuse irradiance near the solar
            disc. (Default: False).

    Returns:
        report: ...
        total: A data collection of total solar irradiance or illuminance in the
            direction of the _srf_azimuth_ and _srf_altitude_.
        direct: A data collection of direct solar irradiance or illuminance in the
            direction of the _srf_azimuth_ and _srf_altitude_.
        diff: A data collection of diffuse sky solar irradiance or illuminance in
            the direction of the _srf_azimuth_ and _srf_altitude_.
        reflect: A data collection of ground reflected solar irradiance or
            illuminance in the direction of the _srf_azimuth_ and _srf_altitude_.
"""

ghenv.Component.Name = 'LB Directional Solar Irradiance'
ghenv.Component.NickName = 'DirSolar'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'


try:
    from ladybug.wea import Wea
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.header import Header
    from ladybug.datatype.illuminance import Illuminance
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

def rad_to_ill(data):
    """Change the data type of an input collection from irradiane to illuminance."""
    head = data.header
    new_header = Header(Illuminance(), 'lux', head.analysis_period, head.metadata)
    return HourlyContinuousCollection(new_header, data.values) if \
        isinstance(data, HourlyContinuousCollection) else \
        data.__class__(new_header, data.values, data.datetimes)


if all_required_inputs(ghenv.Component):
    # set default values
    az = _srf_azimuth_ if _srf_azimuth_ is not None else 180
    alt = _srf_altitude_ if _srf_altitude_ is not None else 0
    gref = _ground_ref_ if _ground_ref_ is not None else 0.2
    isot = not anisotrophic_

    # create the Wea and output irradaince
    wea = Wea(_location, _direct_norm, _diffuse_horiz)
    total, direct, diff, reflect = \
        wea.directional_irradiance(alt, az, gref, isot)
    for dat in (total, direct, diff, reflect):
        dat.header.metadata['altitude'] = alt
        dat.header.metadata['azimuth'] = az

    # convert to illuminace if input data was illuiminance
    if isinstance(_direct_norm.header.data_type, Illuminance):
        total = rad_to_ill(total)
        direct = rad_to_ill(direct)
        diff = rad_to_ill(diff)
        reflect = rad_to_ill(reflect)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Perform a "mass" arithmetic operation between Data Collections. For example,
adding a list of Data Collections into one Data Collection.
-
Note that Data Collections must be aligned in order for this component to run
successfully.
-
Using this component will often be much faster and more elegant compared to
deconstructing the data collection, performing the operation with native
Grasshopper components, and rebuilding the collection.
-

    Args:
        _data: A list of Data Collections to be used in the arithmetic operation.
        _operator_: Text for the operator to use between the Data Collections.
            Valid examples include (+, -, *, /). By default this is + for addition.
        type_: Optional text for a new "type" key in the Data Collection's metadata.
            This will usually show up in most Ladybug visualiztions and it should
            usually change for most types of operations.

    Returns:
        data: A Ladybug data collection object derived from the operation between
            the two data inputs.
"""

ghenv.Component.Name = "LB Mass Arithmetic Operation"
ghenv.Component.NickName = 'MassArithOp'
ghenv.Component.Message = '1.3.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    import ladybug.datatype
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # build the arithmetic statement
    operator = '+' if _operator_ is None else _operator_
    statement = 'data {} data_i'.format(operator)

    # perform the arithmetic operation
    data = _data[0]
    for data_i in _data[1:]:
        data = eval(statement, {'data': data, 'data_i': data_i})  # I love Python!

    # try to replace the data collection type
    try:
        data = data.duplicate()
        if type_:
            data.header.metadata['type'] = type_
        elif 'type' in data.header.metadata:
            d_unit = data.header.unit
            for key in ladybug.datatype.UNITS:
                if d_unit in ladybug.datatype.UNITS[key]:
                    base_type = ladybug.datatype.TYPESDICT[key]()
                    data.header.metadata['type'] = str(base_type)
                    break
            else:
                data.header.metadata['type'] = 'Unknown Data Type'
        if 'System' in data.header.metadata:
            data.header.metadata.pop('System')
        if 'Zone' in data.header.metadata:
            data.header.metadata.pop('Zone')
    except AttributeError:
        pass  # data was not a data collection; just return it anyway


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Calculate humidity metrics from relative humidity, dry bulb temperature and
(if present) atmospheric pressure.
-

    Args:
        _dry_bulb: A data collection representing outdoor dry bulb temperature [C]
        _heat_base_: A number for the base temperature below which a given hour
            is considered to be in heating mode. Default is 18 Celcius, which is
            a common balance point for buildings.
        _cool_base_: A number for the base temperature above which a given hour
            is considered to be in cooling mode. Default is 23 Celcius, which is
            a common balance point for buildings.
    
    Returns:
        hourly_heat: A data collection of heating degree-days.
            Plug this into the 'Time Interval Operation' component to get
            the number of degree-days at different time intervals.
        hourly_cool: A data collection of cooling degree-days.
            Plug this into the 'Time Interval Operation' component to get
            the number of degree-days at different time intervals.
        heat_deg_days: A value indicating the total number of heating degree-days
            over the entire input _dry_bulb collection.
        cool_deg_days: A value indicating the total number of cooling degree-days
            over the entire input _dry_bulb collection.
"""

ghenv.Component.Name = 'LB Degree Days'
ghenv.Component.NickName = 'HDD_CDD'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:
    from ladybug_comfort.degreetime import heating_degree_time, cooling_degree_time
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.datatype.temperaturetime import HeatingDegreeTime, CoolingDegreeTime
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    if _heat_base_ is None:
        _heat_base_ = 18
    if _cool_base_ is None:
        _cool_base_ = 23
    
    hourly_heat = HourlyContinuousCollection.compute_function_aligned(
        heating_degree_time, [_dry_bulb, _heat_base_],
        HeatingDegreeTime(), 'degC-hours')
    hourly_heat.convert_to_unit('degC-days')
    
    hourly_cool = HourlyContinuousCollection.compute_function_aligned(
        cooling_degree_time, [_dry_bulb, _cool_base_],
        CoolingDegreeTime(), 'degC-hours')
    hourly_cool.convert_to_unit('degC-days')
    
    heat_deg_days = hourly_heat.total
    cool_deg_days = hourly_cool.total


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Get statitics of thermal comfort from a Ladybug Comfort Object.
-

    Args:
        _comf_obj: A Ladybug ComfortCollection object from any of the comfort
            model components.
    
    Returns:
        pct_hot: The percent of time that conditions are hotter than acceptable limits.
        pct_neutral: The percent of time that conditions are within acceptable limits
            (aka. the percent of time comfortable).
        pct_cold: The percent of time that conditions are colder than acceptable limits.
"""

ghenv.Component.Name = 'LB Comfort Statistics'
ghenv.Component.NickName = 'ComfStat'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_comfort.collection.base import ComfortCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    assert isinstance(_comf_obj, ComfortCollection), '_comf_obj must be a ' \
        'Ladybug ComfortCollection object. Got {}.'.format(type(_comf_obj))

    pct_hot = _comf_obj.percent_hot
    pct_neutral = _comf_obj.percent_neutral
    pct_cold = _comf_obj.percent_cold


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Convert a hourly Ladybug data collection to a continuous collection at a
specific timestep.
_
This will be done either through linear interpolation or by culling out values
that do not fit the timestep.  It can also be used to convert a discontinous
data collection to a continuous one by linearly interpolating over holes in
the data set.
-

    Args:
        _data: A Ladybug Hourly DataCollection object.  This can be either
            continuous or discontinuous.
        _timestep_: The timestep to which the data will be converted. If this
            is higher than the input _data timestep, values will be
            linerarly interpolated to the new timestep.  If it is lower,
            values that do not fit the timestep will be removed from the
            DataCollection. (Defaut: 1)

    Returns:
        data: A Continuous DataCollection at the input _timestep_.
"""

ghenv.Component.Name = 'LB Convert to Timestep'
ghenv.Component.NickName = 'ToStep'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.datacollection import HourlyDiscontinuousCollection, \
        HourlyContinuousCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check the inputs
    assert isinstance(_data, HourlyDiscontinuousCollection), \
        ' Connected _data is not a Hourly Data Collection. Got{}'.format(type(_data))
    if _timestep_ is not None:
        valid_timesteps = _data.header.analysis_period.VALIDTIMESTEPS
        assert _timestep_ in valid_timesteps, ' Connected _timestep_ is not a'\
            ' hourly timestep.\n Got {}. Choose from: {}'.format(
            _timestep_, sorted(valid_timesteps.keys()))

    # if the data is not continuous, interpolate over holes.
    if not isinstance(_data, HourlyContinuousCollection):
        if _data.validated_a_period is False:
            _data = data.validate_analysis_period
        _data = _data.interpolate_holes()

    # convert the data to the timestep
    if _timestep_ and _timestep_ != _data.header.analysis_period.timestep:
        if _timestep_ < _data.header.analysis_period.timestep:
            data = _data.cull_to_timestep(_timestep_)
        else:
            data = _data.interpolate_to_timestep(_timestep_)
    else:
        data = _data


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create a chart in the Rhino scene with data organized by month.
_
Data will display as a bar chart if the input data is monthly or daily. If the
data is hourly or sub-hourly, it will be plotted with lines and/or a colored
mesh that shows the range of the data within specific percentiles.
-

    Args:
        _data: Data collections (eg. HourlyCollection, MonthlyCollection, or
            DailyCollection), which will be used to generate the monthly chart.
        _base_pt_: An optional Point3D to be used as a starting point to generate
            the geometry of the chart (Default: (0, 0, 0)).
        _x_dim_: An optional number to set the X dimension of each month of the
            chart. (Default: 10 meters).
        _y_dim_: An optional number to set the Y dimension of the entire
            chart (Default: 40 meters).
        stack_: Boolean to note whether multiple connected monthly or daily input
            _data with the same units should be stacked on top of each other.
            Otherwise, all bars for monthly/daily data will be placed next to
            each other.  (Default: False).
        time_marks_: Boolean to note whether the month labels should be replaced with
            marks for the time of day in each month. This is useful for
            displaying hourly data, particularly when the input data is only
            for a month and not the whole year.
        percentile_: An optional number between 0 and 50 to be used for the percentile
            difference from the mean that hourly data meshes display at. For example,
            using 34 will generate hourly data meshes with a range of one standard
            deviation from the mean. Note that this input only has significance when
            the input data collections are hourly. (Default: 34)
        global_title_: A text string to label the entire entire chart.  It will be
            displayed in the lower left of the output chart.  The default will
            display the metadata of the input _data.
        y_axis_title_: A text string to label the Y-axis of the chart.  This can
            also be a list of 2 Y-axis titles if there are two different types
            of data connected to _data and there are two axes labels on either
            side of the chart.  The default will display the data type and
            units of the first (and possibly the second) data collection
            connected to _data.
        legend_par_: An optional LegendParameter object to change the display
            of the chart (Default: None).

    Returns:
        report: ...
        data_mesh: A list of colored meshes that represent the different input data.
            These meshes will resemble a bar chart in the case of monthly or
            daily data and will resemble a band between two ranges for hourly
            and sub-hourly data. Multiple lists of meshes will be output for
            several input data streams.
        data_lines: A list of polylines that represent the input data. These will
            represent the average or total at each hour whenever the input data
            is hourly or monthly-per-hour data.
        col_lines: A list of colored polylines that represent the input data. These
            will only be output when the input data are monthly per hour.
        legend: Geometry representing the legend for the chart, noting which
            colors correspond to which input data.
        borders: A list of lines and polylines representing the axes and intervals
            of the chart.
        labels: A list of text objects that label the borders with month name
            and the intervals of the Y-axis.
        y_title: A text oject for the Y-axis title.
        title: A text object for the global_title.
"""

ghenv.Component.Name = 'LB Monthly Chart'
ghenv.Component.NickName = 'MonthlyChart'
ghenv.Component.Message = '1.3.3'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.monthlychart import MonthlyChart
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_geometry.geometry2d.pointvector import Point2D
    from ladybug_geometry.geometry3d.pointvector import Point3D
    from ladybug_geometry.geometry3d.plane import Plane
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters, tolerance
    from ladybug_rhino.color import color_to_color
    from ladybug_rhino.togeometry import to_point2d
    from ladybug_rhino.fromgeometry import from_mesh2d, from_mesh2d_to_outline, \
        from_polyline2d, from_linesegment2d
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.colorize import ColoredPolyline
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.grasshopper import all_required_inputs, schedule_solution
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and None not in _data:
    # set default values for the chart dimensions
    z_val = _base_pt_.Z if _base_pt_ is not None else 0
    z_val_tol = z_val + tolerance
    _base_pt_ = to_point2d(_base_pt_) if _base_pt_ is not None else Point2D()
    _x_dim_ = _x_dim_ if _x_dim_ is not None else 10.0 / conversion_to_meters()
    _y_dim_ = _y_dim_ if _y_dim_ is not None else 40.0 / conversion_to_meters()
    stack_ = stack_ if stack_ is not None else False
    percentile_ = percentile_ if percentile_ is not None else 34.0
    lpar = legend_par_[0] if len(legend_par_) != 0 else None

    # create the monthly chart object and get the main pieces of geometry
    month_chart = MonthlyChart(_data, lpar, _base_pt_, _x_dim_, _y_dim_,
                               stack_, percentile_)
    if len(legend_par_) > 1:
        if legend_par_[1].min is not None:
            month_chart.set_minimum_by_index(legend_par_[1].min, 1)
        if legend_par_[1].max is not None:
            month_chart.set_maximum_by_index(legend_par_[1].max, 1)

    #  get the main pieces of geometry
    data_lines = []
    d_meshes = month_chart.data_meshes
    if d_meshes is not None:
        data_mesh = [from_mesh2d(msh, z_val_tol) for msh in d_meshes]
        if month_chart.time_interval == 'Monthly':
            data_lines += [l for msh in d_meshes for l in
                           from_mesh2d_to_outline(msh, z_val_tol)]
    d_lines = month_chart.data_polylines
    if d_lines is not None:
        data_lines += [from_polyline2d(lin, z_val_tol) for lin in d_lines]
    borders = [from_polyline2d(month_chart.chart_border, z_val)] + \
            [from_linesegment2d(line, z_val) for line in month_chart.y_axis_lines] + \
            [from_linesegment2d(line, z_val_tol) for line in month_chart.month_lines]
    legend = legend_objects(month_chart.legend)

    # process all of the text-related outputs
    title_txt = month_chart.title_text if global_title_ is None else global_title_
    txt_hgt = month_chart.legend_parameters.text_height
    font = month_chart.legend_parameters.font
    title = text_objects(title_txt, month_chart.lower_title_location, txt_hgt, font)

    # process the first y axis
    y1_txt = month_chart.y_axis_title_text1 if len(y_axis_title_) == 0 else y_axis_title_[0]
    y_title = text_objects(y1_txt, month_chart.y_axis_title_location1, txt_hgt, font)
    if time_marks_:
        txt_h = _x_dim_ / 20 if _x_dim_ / 20 < txt_hgt * 0.75 else txt_hgt * 0.75
        label1 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, z_val)), txt_h, font, 1, 0)
                  for txt, pt in zip(month_chart.time_labels, month_chart.time_label_points)]
        borders.extend([from_linesegment2d(line, z_val_tol) for line in month_chart.time_ticks])
    else:
        label1 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, z_val)), txt_hgt, font, 1, 0)
                  for txt, pt in zip(month_chart.month_labels, month_chart.month_label_points)]
    label2 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, z_val)), txt_hgt, font, 2, 3)
              for txt, pt in zip(month_chart.y_axis_labels1, month_chart.y_axis_label_points1)]
    labels = label1 + label2

    # process the second y axis if it exists
    if month_chart.y_axis_title_text2 is not None:
        y2_txt = month_chart.y_axis_title_text2 if len(y_axis_title_) <= 1 else y_axis_title_[1]
        y_title2 = text_objects(y2_txt, month_chart.y_axis_title_location2, txt_hgt, font)
        y_title = [y_title, y_title2]
        label3 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, z_val)), txt_hgt, font, 0, 3)
                 for txt, pt in zip(month_chart.y_axis_labels2, month_chart.y_axis_label_points2)]
        labels = labels + label3

    # if there are colored lines, then process them to be output from the component
    if month_chart.time_interval == 'MonthlyPerHour':
        cols = [color_to_color(col) for col in month_chart.colors]
        col_lines, month_count = [], len(data_lines) / len(_data)
        for i, pline in enumerate(data_lines):
            col_line = ColoredPolyline(pline)
            col_line.color = cols[int(i / month_count)]
            col_line.thickness = 3
            col_lines.append(col_line)
        # CWM: I don't know why we have to re-schedule the solution but this is the
        # only way I found to get the colored polylines to appear (redraw did not work).
        schedule_solution(ghenv.Component, 2)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>


"""
Create a custom legend for any set of data or range. Creating a legend with this
component allows for a bit more flexibility than what can be achieved by working
with the legends automatically output from different studies.
-

    Args:
        _values: A list of numerical values or data collections that the legend refers
            to. This can also be the minimum and maximum numerical values of the
            data. The legend's maximum and minimum values will be set by the max
            and min of the data set.
        _base_plane_: An optional plane or point to set the location of the
            legend. (Default: Rhino origin - (0, 0, 0))
        title_: A text string representing a legend title. Legends are usually
            titled with the units of the data.
        legend_par_: Optional legend parameters from the  Legend Parameters component.

    Returns:
        mesh: A colored mesh for the legend colors.
        title_obj: A text object for the  legend title.
        label_objs: An array of text objects for the label text.
        label_text: An array of text strings for the label text.
        colors: An array of colors that align with the input _values. This can
            be used to color geometry that aligns with the values.
"""

ghenv.Component.Name = "LB Create Legend"
ghenv.Component.NickName = 'CreateLegend'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.datacollection import BaseCollection
    from ladybug.legend import Legend, LegendParameters
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_plane
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.color import color_to_color
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set default values
    legend_par_ = legend_par_.duplicate() if legend_par_ is not None else \
        LegendParameters()
    if _base_plane_:
        legend_par_.base_plane = to_plane(_base_plane_)
    legend_par_.title = title_

    # create the legend
    values = []
    for val in _values:
        try:
            values.append(float(val))
        except AttributeError:  # assume it's a data collection
            values.extend(val.values)
    legend = Legend(values, legend_par_)

    # separate all of the outputs from this component
    rhino_objs = legend_objects(legend)
    mesh = rhino_objs[0]
    title_obj = rhino_objs[1]
    label_objs = rhino_objs[2:]
    label_text = legend.segment_text
    colors = [color_to_color(col) for col in legend.value_colors]



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Bake a clored mesh into the Rhino scene as a group of colored hatches.
_
This is useful when exporting ladybug graphics from Rhino to vector-based programs
like Inkscape or Illustrator since hatches are exported from Rhino as colored-filled
polygons.
-

    Args:
        _mesh: A colored mesh (or list of colored meshes) to be baked into the Rhino
            scene as groups of colored hatches.
        layer_: Text for the layer name on which the hatch will be added. If unspecified,
            it will be baked onto the currently active layer.
        _run: Set to 'True' to bake the mesh into the scene as hatches.

    Returns:
        report: Reports, errors, warnings ...
"""

ghenv.Component.Name = 'LB Mesh to Hatch'
ghenv.Component.NickName = 'Hatch'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'


try:
    from ladybug_rhino.togeometry import to_mesh3d
    from ladybug_rhino.bakegeometry import add_mesh3d_as_hatch_to_scene
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    lb_mesh = to_mesh3d(_mesh, color_by_face=False)
    add_mesh3d_as_hatch_to_scene(lb_mesh, layer_)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Produce a DDY file with a heating and a cooling design day from an EPW.
_
This method will first check if there is any heating or cooling design day information
contained within the EPW itself. If None is found, the heating and cooling design
days will be derived from analysis of the annual data within the EPW, which is
usually less representative of the climate since only one year of data is used
to derive the DDY (instead of the usual multi-year analysis that is performed for
DDYs distributed with EPWs).
_
Information on the uncertainty introduced by using only one year of data to create
design days can be found in AHSRAE HOF 2013, Chapter 14.14.
-

    Args:
        _epw_file: An .epw file path on your system, from which a .ddy will
            be generated.
        _percentile_: A number between 0 and 50 for the percentile difference
            from the most extreme conditions within the EPW to be used for
            the design day. Typical values are 0.4 and 1.0. (Default: 0.4).
        _folder_: An optional file path to a directory into which the DDY file
            will be written.  If None, the DDY file will be written to the
            ladybug default weather data folder and placed in a sub-folder
            called "ddy".
        _write: Set to "True" to write the .ddy file.

    Returns:
        ddy_file: A .ddy file path that has been written to your system.
"""

ghenv.Component.Name = 'LB EPW to DDY'
ghenv.Component.NickName = 'EPWtoDDY'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

import os

try:
    from ladybug.epw import EPW
    from ladybug.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _write:
    # set default values
    _percentile_ = 0.4 if _percentile_ is None else _percentile_
    _folder_ = os.path.join(folders.default_epw_folder, 'ddy') if _folder_ \
        is None else _folder_
    f_name = os.path.basename(_epw_file).replace('.epw', '.ddy')
    f_path = os.path.join(_folder_, f_name)

    # create the DDY file
    epw = EPW(_epw_file)
    ddy_file = epw.to_ddy(f_path, _percentile_)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Evaluate the percent visibility from geometry to a specific set of points.
_
Such visibility calculations can be used to understand the portions of a building
facade that can see a skyline or landmark when used on the outdoors. When used
on the indoors, they can evaluate the spectator view of a stage, screen, or other
point of interest.
-

    Args:
        _view_points: A list of points that characterize an area of interest to which
            visibility is being evaluated. If the area of interest is more like
            a surface than an individual point, the "LB Generate Point Grid"
            component can be used to obtain a list of points that are evenly
            distributed over the surface.
        pt_weights_: An optional list of numbers that align with the _view_points
            and represent weights of importance for each point.  Weighted
            values should be between 0 and 1 and should be closer to 1 if a
            certain point is more important. The default value for all points
            is 0, which means they all have an equal importance.
        _geometry: Rhino Breps and/or Rhino Meshes for which visibility analysis
            will be conducted. If Breps are input, they will be subdivided using
            the _grid_size to yeild individual points at which analysis will
            occur. If a Mesh is input, visibility analysis analysis will be
            performed for each face of this mesh instead of subdividing it.
        context_: Rhino Breps and/or Rhino Meshes representing context geometry
            that can block visibility from the test _geometry.
        _grid_size: A positive number in Rhino model units for the size of grid
            cells at which the input _geometry will be subdivided for direct sun
            analysis. The smaller the grid size, the higher the resolution of
            the analysis and the longer the calculation will take.  So it is
            recommended that one start with a large value here and decrease
            the value as needed. However, the grid size should usually be
            smaller than the dimensions of the smallest piece of the _geometry
            and context_ in order to yield meaningful results.
        _offset_dist_: A number for the distance to move points from the surfaces
            of the input _geometry.  Typically, this should be a small positive
            number to ensure points are not blocked by the mesh. (Default: 10 cm
            in the equivalent Rhino Model units).
        max_dist_: An optional number to set the maximum distance beyond which the
            end_points are no longer considered visible by the start_points.
            If None, points with an unobstructed view to one another will be
            considered visible no matter how far they are from one another.
        _geo_block_: Set to "True" to count the input _geometry as opaque and
            set to "False" to discount the _geometry from the calculation and
            only look at context_ that blocks the visibility. (Default: True)
        legend_par_: Optional legend parameters from the "LB Legend Parameters"
            that will be used to customize the display of the results.
        _cpu_count_: An integer to set the number of CPUs used in the execution of the
            intersection calculation. If unspecified, it will automatically default
            to one less than the number of CPUs currently available on the
            machine or 1 if only one processor is available.
        _run: Set to "True" to run the component and perform visibility analysis of
            the input _geometry.

    Returns:
        report: ...
        points: The grid of points on the test _geometry that are be used to perform
            the visibility analysis.
        results: A list of numbers that aligns with the points. Each number indicates
            the percentage of the _view_points that are not blocked by context geometry.
        mesh: A colored mesh of the test _geometry representing the percentage of
            the input _geometry's visibility that is not blocked by context.
        legend: A legend showing the number of hours that correspond to the colors
            of the mesh.
        title: A text object for the study title.
        int_mtx: A Matrix object that can be connected to the "LB Deconstruct Matrix"
            component to obtain detailed point-by-point results of the study.
            Each sub-list (aka. branch of the Data Tree) represents one of the
            geometry points used for analysis. The length of each sub-list matches
            the number of _view_points used for the analysis. Each value in the
            sub-list is either a "1", indicating that the vector is visible
            for that vector, or a "0", indicating that the vector is not visible
            for that vector.
"""

ghenv.Component.Name = "LB Visibility Percent"
ghenv.Component.NickName = 'VisibilityPercent'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug.color import Colorset
    from ladybug.graphic import GraphicContainer
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_joined_gridded_mesh3d, to_vector3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.intersect import join_geometry_to_mesh, intersect_mesh_lines
    from ladybug_rhino.grasshopper import all_required_inputs, hide_output, \
        show_output, objectify_output, recommended_processor_count
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
        # set the default values
        _offset_dist_ = _offset_dist_ if _offset_dist_ is not None \
            else 0.1 / conversion_to_meters()
        if pt_weights_:
            assert len(pt_weights_) == len(_view_points), \
                'The number of pt_weights_({}) must match the number of _view_points ' \
                '({}).'.format(len(pt_weights_), len(_view_points))
        workers = _cpu_count_ if _cpu_count_ is not None else recommended_processor_count()

        # create the gridded mesh from the geometry
        study_mesh = to_joined_gridded_mesh3d(_geometry, _grid_size)
        points = [from_point3d(pt.move(vec * _offset_dist_)) for pt, vec in
                  zip(study_mesh.face_centroids, study_mesh.face_normals)]
        hide_output(ghenv.Component, 1)

        # mesh the geometry and context
        shade_mesh = join_geometry_to_mesh(_geometry + context_) if _geo_block_ \
            or _geo_block_ is None else join_geometry_to_mesh(context_)

        # intersect the lines with the mesh
        int_matrix = intersect_mesh_lines(
            shade_mesh, points, _view_points, max_dist_, cpu_count=workers)

        # compute the results
        int_mtx = objectify_output('Visibility Intersection Matrix', int_matrix)
        vec_count = len(_view_points)
        if pt_weights_:  # weight intersections by the input point weights
            tot_wght = sum(pt_weights_) / vec_count
            adj_weights = [wght / tot_wght for wght in pt_weights_]
            results = []
            for int_vals in int_matrix:
                w_res = [ival * wght for ival, wght in zip(int_vals, adj_weights)]
                results.append(sum(w_res) * 100 / vec_count)
        else:  # no need to wieght results
            results = [sum(int_list) * 100 / vec_count for int_list in int_matrix]

        # create the mesh and legend outputs
        graphic = GraphicContainer(results, study_mesh.min, study_mesh.max, legend_par_)
        graphic.legend_parameters.title = '%'
        if legend_par_ is None or legend_par_.are_colors_default:
            graphic.legend_parameters.colors = Colorset.view_study()
        title = text_objects(
            'Visibility Percent', graphic.lower_title_location,
            graphic.legend_parameters.text_height * 1.5,
            graphic.legend_parameters.font)

        # create all of the visual outputs
        study_mesh.colors = graphic.value_colors
        mesh = from_mesh3d(study_mesh)
        legend = legend_objects(graphic.legend)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Convert a DataCollection to the input _to_unit.
-

    Args:
        _data: A DataCollection to be converted to different units.
        to_unit_: Text representing the unit to convert the DataCollection to (eg. m2).
            Connect the _data and see the all_unit output for a list of all
            currently-supported units for a given collection. The default won't
            perform any unit conversion on the output data.

    Returns:
        all_unit: A list of all possible units that the input _data can be converted to.
        data: The converted DataCollection.
"""

ghenv.Component.Name = 'LB To Unit'
ghenv.Component.NickName = 'ToUnit'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    assert isinstance(_data, BaseCollection), \
        '_data must be a Data Collection. Got {}.'.format(type(_data))
    all_unit = _data.header.data_type.units
    data = _data.to_unit(to_unit_) if to_unit_ else _data


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Capture views of the Rhino scene and save them to your hard drive as as a .png files.
_
This is particularly useful when creating animations and one needs to automate
the capturing of views. Note that images will likely have a Rhino world axes icon
in the lower left of the image unless you go to Options > Grid > and uncheck
"Show world axes icon".
-

    Args:
        _file_name: The file name, which the image will be saved as. Note that, for
            animations, each saved image should have a different name. Otherwise,
            the previous image will be overwritten by each successive image.
            Unique names for each animation frame can be achieved by using the
            animating slider to generate the file name.
        _folder_: The folder into which the image file will be written. This should
            be a complete path to the folder. If None, the images will be written
            to one of the following default locations:
            -
            Windows - C:/Users/[USERNAME]/ladybug_tools/resources/captured_views/
            Mac - /Users/[USERNAME]/ladybug_tools/resources/captured_views/
        viewport_: Text for the Rhino viewport name which will be captured. This can
            also be a list of viewports in which case multiple views will be
            captured. If None, the default will be the active viewport (the
            last viewport in which you navigated). Acceptable inputs include:
                -
                Perspective
                Top
                Bottom
                Left
                Right
                Front
                Back
                any view name that has been saved within the Rhino file
        width_: Integer for the width of the image to be captured in pixels. If None,
            the default is the width of the Rhino viewport currently on the screen.
        height_: Integer for the height of the image to be captured in pixels. If None,
            the default is the height of the Rhino viewport currently on the screen.
        mode_: Text for the display mode of the viewport to be captured.If None, the default
            will be the display mode of the active viewport (the last viewport in
            which you navigated). Acceptable inputs include:
                -
                Wireframe
                Shaded
                Rendered
                Ghosted
                X-Ray
                Technical
                Artistic
                Pen
        transparent_: Boolean to note whether the captured .png file should have a
            transparent background. If None or False, the image will have the
            Rhino viewport background color.
        _capture: Set to "True" to capture the image of the Rhino viewport.

    Returns:
        file: The file path of the image taken with this component.

"""
ghenv.Component.Name = 'LB Capture View'
ghenv.Component.NickName = 'CaptureView'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import os

try:
    from ladybug.futil import preparedir
    from ladybug.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
    default_folder = os.path.join(folders.default_simulation_folder, 'captured_views')
except:
    home_folder = os.getenv('HOME') or os.path.expanduser('~')
    default_folder = os.path.join(home_folder, 'captured_views')

try:
    from ladybug_rhino.grasshopper import all_required_inputs, bring_to_front
    from ladybug_rhino.viewport import viewport_by_name, capture_view
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _capture:
    # ensure the component runs last on the canvas
    bring_to_front(ghenv.Component)

    # prepare the folder
    folder = _folder_ if _folder_ is not None else default_folder
    preparedir(folder, remove_content=False)

    # get the viewport objects
    vp_names = viewport_ if len(viewport_) != 0 else [None]
    viewports = [viewport_by_name(vp) for vp in vp_names]

    # save the viewports to images
    for i, view_p in enumerate(viewports):
        f_name = _file_name if len(viewports) == 1 else \
            '{}_{}'.format(_file_name, vp_names[i])
        file_p = os.path.join(folder, f_name)
        fp = capture_view(view_p, file_p, width_, height_, mode_, transparent_)
        print(fp)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>


"""
Load Ladybug data collections from a CSV, JSON, or PKL file.
-

    Args:
        _data_file: A file path to a CSV, JSON or PKL file from which data collections
            will be loaded.
        _load: Set to "True" to load the data collections from the _data_file.

    Returns:
        report: Reports, errors, warnings, etc.
        hb_objs: A list of honeybee objects that have been re-serialized from
            the input file.
"""

ghenv.Component.Name = 'LB Load Data'
ghenv.Component.NickName = 'LoadData'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

try:
    from ladybug.datautil import collections_from_csv, collections_from_json, \
        collections_from_pkl
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import the core ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _load:
    # load the data from the appropriate format
    lower_file = _data_file.lower()
    if lower_file.endswith('.csv'):
        data = collections_from_csv(_data_file)
    elif lower_file.endswith('.json'):
        data = collections_from_json(_data_file)
    elif lower_file.endswith('.pkl'):
        data = collections_from_pkl(_data_file)
    else:
        raise ValueError(
            'Could not recognize the file extension of: {}'.format(_data_file))



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Dragonfly.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Dragonfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
This component updates the Ladybug Tools core libraries and grasshopper components
to either the latest development version available (default) or to a specific
version of the grasshopper plugin.
_
The input version_ does not need to be newer than the current installation and can
be older but grasshopper plugin versions less than 0.3.0 are not supported.
A list of all versions of the Grasshopper plugin and corresponding release notes
can be found at: https://github.com/ladybug-tools/lbt-grasshopper/releases
_
This component can also overwrite the user libraries of standards (constructions,
schedules, modifiers) with a completely fresh copy if clean_standards_ is set to True.
-
    Args:
        _update: Set to True to update your installation of Ladybug Tools to the
            latest development version or to be at the version specified below.
        version_: An optional text string for the version of the grasshopper
            plugin which you would like to update to. The input should contain
            only integers separated by two periods (eg. 1.0.0). The version does
            not need to be newer than the current installation and can be older
            but grasshopper plugin versions less than 0.3.0 are not supported.
            A list of all versions of the Grasshopper plugin can be found
            here - https://github.com/ladybug-tools/lbt-grasshopper/releases
        clean_standards_: Set to True to have the library of standards (constructions,
            schedules, modifiers) overwritten with a completely fresh copy.
            DO NOT SET TO TRUE IF YOU WANT TO KEEP ANY OBJECTS THAT YOU HAVE
            ADDED TO YOUR honeybee_standards FOLDER. If False or None, any
            existing standards will be left alone.
    Returns:
        Vviiiiiz!: !!!
"""

ghenv.Component.Name = 'LB Versioner'
ghenv.Component.NickName = 'Versioner'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '5 :: Version'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.futil import preparedir, nukedir, copy_file_tree, unzip_file
    from ladybug.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders as hb_folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from ladybug_rhino.pythonpath import iron_python_search_path, create_python_package_dir
    from ladybug_rhino.download import download_file_by_name
    from ladybug_rhino.versioning.userobject import UO_FOLDER, GHA_FOLDER
    from ladybug_rhino.versioning.diff import current_userobject_version
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning, \
        give_popup_message
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

import os
import json
import subprocess


def get_gem_directory():
    """Get the directory where measures distributed with Ladybug Tools are installed."""
    measure_folder = os.path.join(folders.ladybug_tools_folder, 'resources', 'measures')
    if not os.path.isdir(measure_folder):
        os.makedirs(measure_folder)
    return measure_folder


def get_standards_directory():
    """Get the directory where Honeybee standards are installed."""
    hb_folder = os.path.join(folders.ladybug_tools_folder, 'resources', 'standards')
    if not os.path.isdir(hb_folder):
        os.makedirs(hb_folder)
    return hb_folder


def remove_dist_info_files(directory):
    """Remove all of the PyPI .dist-info folders from a given directory.
    
    Args:
        directory: A directory containing .dist-info folders to delete.
    """
    for fold in os.listdir(directory):
        if fold.endswith('.dist-info'):
            nukedir(os.path.join(directory, fold), rmdir=True)


def get_config_dict():
    """Get a dictionary of the ladybug configurations.

    This is needed in order to put the configurations back after update.
    """
    with open(folders.config_file, 'r') as cfg:
        config_dict = json.load(cfg)
    return config_dict


def set_config_dict(config_dict):
    """Set the configurations using a dictionary.

    Args:
        config_dict: A dictionary of configuration paths.
    """
    with open(folders.config_file, 'w') as fp:
        json.dump(config_dict, fp, indent=4)


def update_libraries_pip(python_exe, package_name, version=None, target=None):
    """Change python libraries to be of a specific version using pip.
    Args:
        python_exe: The path to the Python executable to be used for installation.
        package_name: The name of the PyPI package to install.
        version: An optional string for the version of the package to install.
            If None, the library will be updated to the latest version with -U.
        target: An optional target directory into which the package will be installed.
        """
    # build up the command using the inputs
    if version is not None:
        package_name = '{}=={}'.format(package_name, version)
    cmds = [python_exe, '-m', 'pip', 'install', package_name]
    if version is None:
        cmds.append('-U')
    if target is not None:
        cmds.extend(['--target', target, '--upgrade'])

    # execute the command and print any errors
    print('Installing "{}" version via pip'.format(package_name))
    use_shell = True if os.name == 'nt' else False
    process = subprocess.Popen(
        cmds, shell=use_shell, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output = process.communicate()
    stdout, stderr = output
    error_msg = 'Package "{}" may not have been updated correctly\n' \
        'or its usage in the plugin may have changed. See pip stderr below:\n' \
        '{}'.format(package_name, stderr)
    return error_msg


def download_repo_github(repo, target_directory, version=None):
    """Download a repo of a particular version from from github.
    Args:
        repo: The name of a repo to be downloaded (eg. 'lbt-grasshopper').
        target_directory: the directory where the library should be downloaded to.
        version: The version of the repository to download. If None, the most
            recent version will be downloaded. (Default: None)
        """
    # download files
    if version is None:
        url = "https://github.com/ladybug-tools/{}/archive/master.zip".format(repo)
    else:
        url = "https://github.com/ladybug-tools/{}/archive/v{}.zip".format(repo, version)
    zip_file = os.path.join(target_directory, '%s.zip' % repo)
    print ('Downloading "{}"  github repository to: {}'.format(repo, target_directory))
    download_file_by_name(url, target_directory, zip_file)

    #unzip the file
    unzip_file(zip_file, target_directory)

    # try to clean up the downloaded zip file
    try:
        os.remove(zip_file)
    except:
        print ('Failed to remove downloaded zip file: {}.'.format(zip_file))

    # return the directory where the unzipped files live
    if version is None:
        return os.path.join(target_directory, '{}-master'.format(repo))
    else:
        return os.path.join(target_directory, '{}-{}'.format(repo, version))


def parse_lbt_gh_versions(lbt_gh_folder):
    """Parse versions of compatible libs from a clone of the lbt-grasshopper repo.
    Args:
        lbt_gh_folder: Path to the clone of the lbt-grasshopper repo

    Returns:
        A dictionary of library versions formatted like so (but with actual version
        numbers in place of '0.0.0').

        {
            'lbt-dragonfly' = '0.0.0',
            'ladybug-rhino' = '0.0.0',
            'lbt-recipes' = '0.0.0',
            'honeybee-openstudio-gem' = '0.0.0',
            'lbt-measures' = '0.0.0',
            'honeybee-standards' = '0.0.0',
            'honeybee-energy-standards' = '0.0.0',
            'ladybug-grasshopper': '0.0.0',
            'honeybee-grasshopper-core': '0.0.0',
            'honeybee-grasshopper-radiance': '0.0.0',
            'honeybee-grasshopper-energy': '0.0.0',
            'dragonfly-grasshopper': '0.0.0',
            'ladybug-grasshopper-dotnet': '0.0.0'
        }
    """
    # set the names of the libraries to collect and the version dict
    version_dict = {
        'lbt-dragonfly': None,
        'ladybug-rhino': None,
        'lbt-recipes': None,
        'honeybee-standards': None,
        'honeybee-energy-standards': None,
        'honeybee-openstudio-gem': None,
        'lbt-measures': None,
        'ladybug-grasshopper': None,
        'honeybee-grasshopper-core': None,
        'honeybee-grasshopper-radiance': None,
        'honeybee-grasshopper-energy': None,
        'dragonfly-grasshopper': None,
        'ladybug-grasshopper-dotnet': None
        }
    libs_to_collect = list(version_dict.keys())

    def search_versions(version_file):
        """Search for version numbers within a .txt file."""
        with open(version_file) as ver_file:
            for row in ver_file:
                try:
                    library, version = row.strip().split('==')
                    if library in libs_to_collect:
                        version_dict[library] = version
                except Exception :  # not a row with a ladybug tools library
                    pass

    # search files for versions
    requirements = os.path.join(lbt_gh_folder, 'requirements.txt')
    dev_requirements = os.path.join(lbt_gh_folder, 'dev-requirements.txt')
    ruby_requirements = os.path.join(lbt_gh_folder, 'ruby-requirements.txt')
    search_versions(requirements)
    search_versions(dev_requirements)
    search_versions(ruby_requirements)
    return version_dict


if all_required_inputs(ghenv.Component) and _update is True:
    # ensure that Python has been installed in the ladybug_tools folder
    py_exe, py_lib = hb_folders.python_exe_path, hb_folders.python_package_path
    assert py_exe is not None and py_lib is not None, \
        'No valid Python instalation was found at: {}.\nThis is a requirement in ' \
        'order to contine with installation'.format(
            os.path.join(folders.ladybug_tools_folder, 'python'))

    # get the compatiable versions of all the dependencies
    temp_folder = os.path.join(folders.ladybug_tools_folder, 'temp')
    lbt_gh_folder = download_repo_github('lbt-grasshopper', temp_folder, version_)
    ver_dict = parse_lbt_gh_versions(lbt_gh_folder)
    ver_dict['lbt-grasshopper'] = version_

    # install the core libraries
    print ('Installing Ladybug Tools core Python libraries.')
    config_dict = get_config_dict() # load configurations so they can be put back after update
    df_ver = ver_dict['lbt-dragonfly']
    stderr = update_libraries_pip(py_exe, 'lbt-dragonfly[cli]', df_ver)
    if os.path.isdir(os.path.join(py_lib, 'lbt_dragonfly-{}.dist-info'.format(df_ver))):
        print ('Ladybug Tools core Python libraries successfully installed!\n ')
    else:
        give_warning(ghenv.Component, stderr)
        print (stderr)
    set_config_dict(config_dict)  # restore the previous configurations

    # install the library needed for interaction with Rhino
    print ('Installing ladybug-rhino Python library.')
    rh_ver = ver_dict['ladybug-rhino']
    stderr = update_libraries_pip(py_exe, 'ladybug-rhino[cli]', rh_ver)
    if os.path.isdir(os.path.join(py_lib, 'ladybug_rhino-{}.dist-info'.format(rh_ver))):
        print ('Ladybug-rhino Python library successfully installed!\n ')
    else:
        give_warning(ghenv.Component, stderr)
        print (stderr)
    if os.name != 'nt':  # make sure libraries are copied to the rhino scripts folder
        iron_python_search_path(create_python_package_dir())

    # install the grasshopper components
    print ('Installing Ladybug Tools Grasshopper components.')
    gh_ver = ver_dict['lbt-grasshopper']
    stderr = update_libraries_pip(py_exe, 'lbt-grasshopper', gh_ver, UO_FOLDER)
    lbgh_ver = ver_dict['ladybug-grasshopper']
    if os.path.isdir(os.path.join(UO_FOLDER, 'ladybug_grasshopper-{}.dist-info'.format(lbgh_ver))):
        print ('Ladybug Tools Grasshopper components successfully installed!\n ')
        remove_dist_info_files(UO_FOLDER)  # remove the .dist-info files
    else:
        give_warning(ghenv.Component, stderr)
        print (stderr)

    # install the .gha Grasshopper components
    gha_location = os.path.join(GHA_FOLDER, 'ladybug_grasshopper_dotnet')
    if os.path.isdir(gha_location):
        msg = '.gha files already exist in your Components folder and cannot be ' \
            'deleted while Grasshopper is open.\nClose Grasshopper, delete the ' \
            'ladybug_grasshopper_dotnet folder at\n{}\nand rerun this versioner ' \
            'component to get the new .gha files.\nOr simply keep '\
            'using the old .gha component if you do not need the latest ' \
            '.gha features.\n '.format(gha_location)
        print (msg)
    else:
        gha_ver = ver_dict['ladybug-grasshopper-dotnet']
        stderr = update_libraries_pip(py_exe, 'ladybug-grasshopper-dotnet', gha_ver, GHA_FOLDER)
        package_dir = os.path.join(
            GHA_FOLDER, 'ladybug_grasshopper_dotnet-{}.dist-info'.format(gha_ver))
        if os.path.isdir(package_dir):
            print ('Ladybug Tools .gha Grasshopper components successfully installed!\n ')
            remove_dist_info_files(GHA_FOLDER)  # remove the dist-info files
        else:
            give_warning(ghenv.Component, stderr)
            print (stderr)

    # install the lbt_recipes package
    print ('Installing Ladyubug Tools Recipes.')
    rec_ver = ver_dict['lbt-recipes']
    stderr = update_libraries_pip(py_exe, 'lbt-recipes', rec_ver)
    if os.path.isdir(os.path.join(py_lib, 'lbt_recipes-{}.dist-info'.format(rec_ver))):
        print ('Ladyubug Tools Recipes successfully installed!\n ')
    else:
        give_warning(ghenv.Component, stderr)
        print (stderr)

    # install the honeybee-openstudio ruby gem
    gem_ver = ver_dict['honeybee-openstudio-gem']
    print ('Installing Honeybee-OpenStudio gem version {}.'.format(gem_ver))
    gem_dir = get_gem_directory()
    base_folder = download_repo_github('honeybee-openstudio-gem', gem_dir, gem_ver)
    source_folder = os.path.join(base_folder, 'lib')
    lib_folder = os.path.join(gem_dir, 'honeybee_openstudio_gem', 'lib')
    print ('Copying "honeybee_openstudio_gem" source code to {}\n '.format(lib_folder))
    copy_file_tree(source_folder, lib_folder)
    nukedir(base_folder, True)

    # install the lbt-measures ruby gem
    mea_ver = ver_dict['lbt-measures']
    print ('Installing Ladybug Tools Measures version {}.'.format(mea_ver))
    base_folder = download_repo_github('lbt-measures', gem_dir, mea_ver)
    source_folder = os.path.join(base_folder, 'lib')
    print ('Copying "lbt_measures" source code to {}\n '.format(gem_dir))
    copy_file_tree(source_folder, gem_dir)
    nukedir(base_folder, True)

    # always update the honeybee-energy-standards package
    print ('Installing Honeybee energy standards.')
    stand_dir = get_standards_directory()
    hes_ver = ver_dict['honeybee-energy-standards']
    if os.path.isdir(os.path.join(stand_dir, 'honeybee_energy_standards')):
        nukedir(os.path.join(stand_dir, 'honeybee_energy_standards'), True)
    stderr = update_libraries_pip(py_exe, 'honeybee-energy-standards', hes_ver, stand_dir)
    if os.path.isdir(os.path.join(stand_dir, 'honeybee_energy_standards-{}.dist-info'.format(hes_ver))):
        print ('Honeybee energy standards successfully installed!\n ')
        remove_dist_info_files(stand_dir)  # remove the dist-info files
    else:
        give_warning(ghenv.Component, stderr)
        print (stderr)

    # install the standards libraries if requested or if they don't exist
    if clean_standards_ or not os.path.isdir(os.path.join(stand_dir, 'honeybee_standards')):
        print ('Installing Ladybug Tools standards libraries (constructions, schedules, etc.).')
        hs_ver = ver_dict['honeybee-standards']
        if os.path.isdir(os.path.join(stand_dir, 'honeybee_standards')):
            nukedir(os.path.join(stand_dir, 'honeybee_standards'), True)
        stderr = update_libraries_pip(py_exe, 'honeybee-standards', hs_ver, stand_dir)
        if os.path.isdir(os.path.join(stand_dir, 'honeybee_standards-{}.dist-info'.format(hs_ver))):
            print ('Honeybee standards successfully installed!\n ')
            remove_dist_info_files(stand_dir)  # remove the dist-info files
        else:
            give_warning(ghenv.Component, stderr)
            print (stderr)

    # delete the temp folder and give a completion message
    nukedir(temp_folder, True)
    version = 'LATEST' if version_ is None else version_
    success_msg = 'Change to Version {} Successful!'.format(version)
    restart_msg = 'RESTART RHINO to load the new components + library.'
    sync_msg = 'The "LB Sync Grasshopper File" component can be used\n' \
        'to sync Grasshopper definitions with your new installation.'
    for msg in (success_msg, restart_msg, sync_msg):
        print (msg)
    give_popup_message('\n'.join([restart_msg, sync_msg]), success_msg)

    # do a check to see if the versioner has, itself, been updated
    new_version = current_userobject_version(ghenv.Component)
    current_version = ghenv.Component.Message
    if new_version != current_version:
        msg = 'The Versioner component has, itself, been changed between the\n' \
            'current version ({}) and the version you are changing to ({}).\n' \
            'It is recommended that you resart Rhino and run the new Versioner\n' \
            'coponent to ensure that everything is consistent.'.format(
                current_version, new_version)
        print (msg)
        give_warning(ghenv.Component, msg)
else:  # give a message to the user about what to do
    print ('Make sure you are connected to the internet and set _update to True!')



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Apply an analysis period to a data collection.
-

    Args:
        _data: A Ladybug data collection object.
        _period: A Ladybug analysis period from the "LB AnalysisPeriod" component.

    Returns:
        data: The data collection with the analysis period applied to it.
"""

ghenv.Component.Name = 'LB Apply Analysis Period'
ghenv.Component.NickName = 'ApplyPer'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from ladybug.analysisperiod import AnalysisPeriod
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    assert isinstance(_data, BaseCollection), \
        '_data must be a Data Collection. Got {}.'.format(type(_data))
    assert isinstance(_period, AnalysisPeriod), '_period must be an Analysis' \
        ' Period. Got {}.'.format(type(_period))
    data = _data.filter_by_analysis_period(_period)


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>


"""
Sync the Ladybug Tools components in a Grasshopper file with the version of the
components that currently exist in the Grasshopper toolbar.
-
This is useful for updating old Grasshopper definitions to newer plugin versions.
However, this component will sync components regardless of version number or
date, even if the components in the toolbar are of an older version than those
currently on the Grasshopper canvass.
-
Any components that cannot be updated automatically (because their inputs or
outputs have changed) will be circled in red and should be replaced manually.
-

    Args:
        _sync: Set to "True" to have this component to search through the
            current Grasshopper file and sync all Ladybug Tools components
            with the version in the Grasshopper toolbar.
    
    Returns:
        report: Errors, warnings, etc.
"""

ghenv.Component.Name = 'LB Sync Grasshopper File'
ghenv.Component.NickName = 'SyncGHFile'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '5 :: Version'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug_rhino.versioning.gather import gather_canvas_components
    from ladybug_rhino.versioning.diff import sync_component
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _sync:
    # load all of the GHPython userobjects and update the versions
    components = gather_canvas_components(ghenv.Component)
    report_init = []
    for comp in components:
        try:
            report_init.append(sync_component(comp, ghenv.Component))
        except Exception:
            if hasattr(comp, 'Name'):
                msg = 'Failed to Update "{}"'.format(comp.Name)
                print msg
                give_warning(ghenv.Component, msg)
    report = '\n'.join(r for r in report_init if r)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Compute Incident Radiation values for any sky matrix in real time using the Geometry/Sky
intersection matrix produced by the "LB Incident Radiation" component.
_
Using this component enables one to scroll through radiation on an hour-by-hour
or month-by-month basis in a manner that is an order of magnitude faster than
running each sky matrix through the "LB Incident Radiation" component.
_
The speed of this component is thanks to the fact that the Geometry/Sky intersection
matrix contains the relationship between the geometry and each patch of the sky.
So computing new radiation values is as simple as multiplying the sky matrix by
the intersection matrix.
-

    Args:
        _int_mtx: A Geometry/Sky Intersection Matrix from the "LB Incident Radiation" 
            component. This matrix contains the relationship between each point of
            the analyzed geometry and each patch of the sky.
        _sky_mtx: A Sky Matrix from the "LB Cumulative Sky Matrix" component, which
            describes the radiation coming from the various patches of the sky.
            The "LB Sky Dome" component can be used to visualize any sky matrix.

    Returns:
        results: A list of numbers that aligns with the points of the original analysis
            performed with the "LB Incident Radiation"  component. Each number
            indicates the cumulative incident radiation received by each of the
            points from the sky matrix in kWh/m2. To visualize these radiation
            values in the Rhino scene, connect these values to the "LB Spatial
            Heatmap" component along with the mesh output from the original
            analysis with the "LB Incident Radiation"  component.
"""

ghenv.Component.Name = 'LB Real Time Incident Radiation'
ghenv.Component.NickName = 'RTrad'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '0'


try:  # python 2
    from itertools import izip as zip
except ImportError:  # python 3
    pass

try:
    from ladybug_rhino.grasshopper import all_required_inputs, de_objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # deconstruct the matrices and get the total radiation from each patch
    int_mtx = de_objectify_output(_int_mtx)
    sky_mtx = de_objectify_output(_sky_mtx)
    total_sky_rad = [dirr + difr for dirr, difr in zip(sky_mtx[1], sky_mtx[2])]
    ground_rad = [(sum(total_sky_rad) / len(total_sky_rad)) * sky_mtx[0][1]] * len(total_sky_rad)
    all_rad = total_sky_rad + ground_rad 

    # compute the results
    results = []
    for pt_rel in int_mtx:
        results.append(sum(r * w for r, w in zip(pt_rel, all_rad)))
